// SNES Myth Menu
// C version 0.20 WIP 4
//
// Mic, 2010

#include "snes.h"
#include "dma.h"
#include "lzss_decode.h"
#include "myth.h"
#include "neo2.h"


// The default header that comes with SNESC specifies a LOROM configuration. Remove this define if HIROM is used.
//#define LOROM

#ifdef LOROM
#define GAME_LIST_BANK 1
#else
#define GAME_LIST_BANK 0
#endif

#define NUMBER_OF_GAMES_TO_SHOW 9

// For use with can_games_list_scroll
typedef enum
{
	DIRECTION_UP,
	DIRECTION_DOWN
} scrollDirection_t;

typedef enum
{
	SORT_LOGICALLY,
	SORT_ALPHABETICALLY
} sortOrder_t;

typedef struct
{
	u8 romSize;
	u8 romZ;
	char copyStr[2];
	u8 countLo, countHi;
} psramParams_t;


const psramParams_t psramParameters[8] =
{
	{0x04, 0x3c, "05", 0x04, 0x00},	//  4M
	{0x08, 0x38, "09", 0x08, 0x00},	//  8M
	{0x09, 0x30, "17", 0x08, 0x01},	// 16M
	{0x0a, 0x20, "25", 0x08, 0x02},	// 24M
	{0x0b, 0x20, "33", 0x08, 0x03},	// 32M
	{0x0c, 0x00, "41", 0x08, 0x04},	// 40M
	{0x0d, 0x00, "49", 0x08, 0x05},	// 48M
	{0x0e, 0x00, "65", 0x08, 0x07}	// 64M
};



// Some resources defined in separate source files
extern char bg_patterns, bg_patterns_end;
extern char bg_palette;
extern char bg_map, bg_map_end;

extern u8 font[];


u8 cardModel, cpID;
u16 romAddressPins;
u8 romSize, sramSize;
sortOrder_t sortOrder = SORT_LOGICALLY;
char sortLetter = 'A';

struct
{
	u16 count;
	u16 firstShown;
	u16 highlighted;
} gamesList;

struct
{
	u16 firstLetterIdx[27];
	u16 alphaToLogical[500];
	u8 visited[500];
} gbaCardAlphabeticalIdx;

// All the text strings are written to this buffer and then sent to VRAM using DMA in order to
// improve performance.
//
char bg0Buffer[0x800];
u8 bg0BufferDirty;

// These three strings are modified at runtime by the menu, so they are declared separately in order to get them into
// the .data section rather than .rodata.
//
char MS2[] = "\xff\x15\x02\x02 LOADING......(  )";
char MS3[] = "\xff\x0b\x02\x02                        \xff\x09\x02\x00SECONDARY CART:";
char MS4[] = "\xff\x15\x02\x02 GAME (001)";


// Each of these metastrings are onm the format 0xff,row,column,palette,"actual text". A single metastring can go on
// indefinitely until a null-terminator is reached.
//
char *metaStrings[] =
{
	"\xff\x03\x01\x00 MENU V 0.20\xff\x06\x02\x00GAMES\xff\x17\x03\x03\x42\xff\x17\x04\x00: RUN, \xff\x17\x0c\x03Y \
     \xff\x17\x0d\x00: RUN 2ND CART\xff\x02\x01\x03 NEO POWER SNES MYTH CARD (A)\xff\x1a\x04\x05\x22 2010 WWW.NEOFLASH.COM     ",
	"\xff\x01\xfe\x0f\x0a\x68\x69\x6A\x20\x71\x72\x73\x20\x7a\x7b\x7c\x83\x84\x85\xfe\x10\x0a\x6b\x6c\x6d\x20\x74\x75 \
	 \x76\x20\x7d\x7e\x7f\x86\x87\x88xfe\x11\x0a\x6e\x6f\x70\x20\x77\x78\x79\x20\x80\x81\x82\x89\x8a\x8b\xfe\x17\x06 \
	 \x06\xff\x04                             ",
	MS2,
	MS3,
	MS4,
	"\xff\x02\x1b\x03(B)",
	"\xff\x02\x1b\x03(C)",
	"\xff\x03\x14\x00(HARD 0.1)",
	"\xff\x03\x14\x00(HARD 0.1)",
	"\xff\x03\x14\x00(HARD 0.1)",
    "\xff\x16\x03\x02          ",
    "\xff\x16\x03\x02SAVE 2KB  ",
    "\xff\x16\x03\x02SAVE 8KB  ",
    "\xff\x16\x03\x02SAVE 32KB ",
    "\xff\x16\x03\x02SAVE 64KB ",
    "\xff\x16\x03\x02SAVE 128KB",
    "\xff\x15\x01\x02 DSP      ",
	" ",
	" ",
	" ",
	"\xff\x09\x14\x05 8M/64M",
	"\xff\x09\x14\x05\x31\x36M/64M",
	"\xff\x09\x14\x05\x32\x34M/64M",
	"\xff\x09\x14\x05\x33\x32M/64M",
	"\xff\x09\x14\x05\x34\x30M/64M",
	"\xff\x09\x14\x05\x34\x38M/64M",
	"\xff\x09\x14\x05\x35\x36M/64M",
	"\xff\x09\x14\x05\x36\x34M/64M",
	"\xff\x0a\x01\x05 A24 & A25 TEST OK !  ",
	" ",
	"\xff\x09\x01\x02 TESTING PSRAM... ",
	"\xff\x0b\x01\x03 PSRAM TEST OK    ",
	"\xff\x0b\x01\x01 PSRAM TEST ERROR!",
	"\xff\x08\x10\x00\x66",		// Up arrow
	"\xff\x12\x10\x00\x67",		// Down arrow
	"\xff\x08\x10\x00 ",		// Up arrow clear
	"\xff\x12\x10\x00 ",		// Down arrow clear
	// Game sizes (strings nbr 37-47)
	"\xff\x15\x0f\x02\x34M ",
	"\xff\x15\x0f\x02?M ",
	"\xff\x15\x0f\x02?M ",
	"\xff\x15\x0f\x02?M ",
	"\xff\x15\x0f\x02\x38M ",
	"\xff\x15\x0f\x02\x31\x36M",
	"\xff\x15\x0f\x02\x32\x34M",
	"\xff\x15\x0f\x02\x33\x32M",
	"\xff\x15\x0f\x02\x34\x30M",
	"\xff\x15\x0f\x02\x34\x38M",
	"\xff\x15\x0f\x02\x36\x34M",
	// ROM types
	"\xff\x15\x13\x02HIROM",
	"\xff\x15\x13\x02LOROM",
	// DSP types
	"\xff\x15\x19\x02    ",
	"\xff\x15\x19\x02\x44SP1",
	"\xff\x15\x19\x02\x44SP2",
	"\xff\x15\x19\x02\x44SP3",
	"\xff\x15\x19\x02\x44SP4",
	"\xff\x15\x19\x02\x44SP5",
	"\xff\x15\x19\x02\x44SP6",
	"\xff\x15\x19\x02SFX ",
	"\xff\x03\x01\x00 MENU V 0.20\xff\x06\x02\x00GAMES\xff\x17\x03\x03Y\xff\x17\x04\x00: RUN, \xff\x17\x0c\x03\x42 \
     \xff\x17\x0d\x00: GO BACK\xff\x02\x01\x03 NEO POWER SNES MYTH CARD (A)\xff\x1a\x04\x05\x22 2010 WWW.NEOFLASH.COM     \
     \xff\x0d\x01\x00 SYSTEM INFO:",
	// Regions (59-60)
    "\xff\x0f\x01\x02 NTSC (60HZ)",
    "\xff\x0f\x01\x02 PAL  (50HZ)",
	// CPU versions (61-64)
    "\xff\x10\x01\x02 S-CPU  V0",
    "\xff\x10\x01\x02 S-CPU  V1",
    "\xff\x10\x01\x02 S-CPU  V2",
    "\xff\x10\x01\x02 S-CPU  V3",
    // PPU1 versions
    "\xff\x11\x01\x02 S-PPU1 V0",
    "\xff\x11\x01\x02 S-PPU1 V1",
    "\xff\x11\x01\x02 S-PPU1 V2",
    "\xff\x11\x01\x02 S-PPU1 V3",
    // PPU2 versions
    "\xff\x12\x01\x02 S-PPU2 V0",
    "\xff\x12\x01\x02 S-PPU2 V1",
    "\xff\x12\x01\x02 S-PPU2 V2",
    "\xff\x12\x01\x02 S-PPU2 V3"
};

const u8 ppuRegData1[12] =
{
     3,		// OBJ data select
	 0,		// OAM address low
	 0,		// OAM address high
	 0,		// OAM data
	 1,		// BG mode
	 0,		// Mosaic (off)
	 0x20,	// BG0 name table size/address
	 0x30,	// BG1 name table size/address
	 0,		// BG2 name table size/address
	 0,		// BG3 name table size/address
	 0x40,	// BG0/1 pattern table address
	 0		// BG2/3 pattern table address
};

const u8 ppuRegData2[9] =
{
     3,		// Enable BG0 and BG1
	 0,		// Sub screen
	 0,		// Window mask
	 0,		// Window mask
	 2,		// Color addition setting
	 0x20,	// Color addition setting
	 0,
	 0,
	 0
};

const u16 fontColors[] =
{
	0x7fff, 0x3c7f, 0x7fff, 0x47f1,
	0x1fc6, 0x7f18, 0x31ed, 0x718f
};



void wait_nmi()
{
	while (REG_RDNMI & 0x80);
	while (!(REG_RDNMI & 0x80));
}


// Set the value of a pointer to a full 24-bit bank:offset pair
//
// E.g. set_full_pointer((void**)&a_pointer, 0x7f, 0x8000) will make
// a_pointer point to 0x7f8000.
//
void set_full_pointer(void **pptr, u8 bank, u16 offset)
{
	u8 *bp = (u8*)pptr;
	u16 *wp = (u16*)pptr;

	bp[2] = bank;
	*wp = offset;
}


void dma_bg0_buffer()
{
	wait_nmi();
	load_vram(bg0Buffer, 0x2000, 0x800);
	bg0BufferDirty = 0;
}


void clear_screen()
{
	int i;

	for (i = 0; i < 0x400; i++)
	{
		bg0Buffer[i + i] = ' ';
		bg0Buffer[i + i + 1] = 0;
	}
	bg0BufferDirty = 1;
}


// Return an unsigned short that contains the status of all 8 buttons and the dpad
// Note that this function is blocking.
//
u16 read_joypad()
{
	int i;
	u8 b, lo, hi;
	u16 retVal = 0;

	for (;;)
	{
		// Delay
		for (i = 0; i < 32; i++);

		REG_NMI_TIMEN = 1;

		while (!((b = REG_HVB_JOY) & 1));
		while ((b = REG_HVB_JOY) & 1);
		lo = REG_JOY1L;
		hi = REG_JOY1H;

		retVal &= 0xff00;
		retVal |= lo;
		if (!lo)
		{
			retVal &= 0x00ff;
			retVal |= hi << 8;
			if (!hi)
			{
				continue;
			}
		}

		b = 1;
		while (b)
		{
			REG_NMI_TIMEN = 1;
			while (!((b = REG_HVB_JOY) & 1));
			while ((b = REG_HVB_JOY) & 1);
			b = REG_JOY1L | REG_JOY1H;
		}

		break;
	}

	return retVal;
}


// Update the "GAME (nnn)" string.
// E.g. update_game_number_string(-2) to decrease nnn by 2.
//
void update_game_number_string(int offset)
{
	if (offset > 0)
	{
		for (; offset > 0; offset--)
		{
			MS4[0xd]++;
			if (MS4[0xd] > '9')
			{
				MS4[0xd] = '0';
				MS4[0xc]++;
				if (MS4[0xc] > '9')
				{
					MS4[0xc] = '0';
					MS4[0xb]++;
				}
			}
		}
	}
	else if (offset < 0)
	{
		offset = -offset;
		for (; offset > 0; offset--)
		{
			MS4[0xd]--;
			if (MS4[0xd] < '0')
			{
				MS4[0xd] = '9';
				MS4[0xc]--;
				if (MS4[0xc] < '0')
				{
					MS4[0xc] = '9';
					MS4[0xb]--;
				}
			}
		}
	}
}


// Convert 1-bit font data to 4-bit (the three remaining bitplanes are all zeroed).
//
void expand_font_data()
{
	int i, j;
	u8 *bp = font;

	REG_VRAM_ADDR_L = 0x00;
	REG_VRAM_ADDR_H = 0x02;
	REG_VRAM_INC = VRAM_WORD_ACCESS;

	for (i = 0; i < 0xd2; i++)
	{
		// Lower two planes
		for (j = 0; j < 8; j++)
		{
			REG_VRAM_DATAW1 = *bp++;
			REG_VRAM_DATAW2 = 0;
		}
		// Upper two planes
		for (j = 0; j < 8; j++)
		{
			REG_VRAM_DATAW1 = 0;
			REG_VRAM_DATAW2 = 0;
		}
	}
}


void load_font_colors()
{
	u8 i;

	for (i = 0; i < 8; i++)
	{
		REG_CGRAM_ADDR = (i << 4) + 1;
		REG_CGRAM_DATAW = (u8)fontColors[i];
		REG_CGRAM_DATAW = fontColors[i] >> 8;
	}
}


// Return the number of games stored on the GBA card
//
u16 count_games_on_gba_card()
{
	u16 cnt;
	u8 *pGames;

	set_full_pointer((void**)&pGames, GAME_LIST_BANK, 0xc800);

	// Loop until a chunk is found that doesn't begin with the value 0xff
	for (cnt = 0; *pGames == 0xff; pGames += 0x40)
	{
		if (++cnt == 500)
		{
			break;
		}
	}

	return cnt;
}



void swap(u16 *a, u16 *b)
{
  	int t=*a; *a=*b; *b=t;
}

// Quicksort implementation taken from Wikipedia, with some modifications
//
void quick_sort(u16 *arr, int beg, int end, int (*cmp)(u16,u16))
{
	int piv, l, r;

	if (end > beg + 1)
  	{
    	piv = arr[beg], l = beg + 1, r = end;
    	while (l < r)
    	{
			if (cmp(arr[l], piv) <= 0)
			{
				l++;
			}
			else
			{
				swap(&arr[l], &arr[--r]);
			}
		}
    	swap(&arr[--l], &arr[beg]);
    	quick_sort(arr, beg, l, cmp);
    	quick_sort(arr, r, end, cmp);
  	}
}

// Game title string comparator used when sorting
//
int cmp_game_titles(u16 a, u16 b)
{
	u8 *pGameA, *pGameB;

	set_full_pointer((void**)&pGameA, GAME_LIST_BANK, 0xc800 + (a << 6));
	set_full_pointer((void**)&pGameB, GAME_LIST_BANK, 0xc800 + (b << 6));

	return strcmp(&pGameA[0xc], &pGameB[0xc]);
}


void create_alphabetical_index()
{
	int i;
	char c;
	u8 *pGames;

	set_full_pointer((void**)&pGames, GAME_LIST_BANK, 0xc800);

	for (i = 0; i < gamesList.count; i++)
	{
		gbaCardAlphabeticalIdx.alphaToLogical[i] = i;
	}

	quick_sort(gbaCardAlphabeticalIdx.alphaToLogical, 0, gamesList.count - 1, cmp_game_titles);
}


// Returns non-zero if the games list can scroll in the given direction
//
int can_games_list_scroll(scrollDirection_t direction)
{
	if ((direction == DIRECTION_DOWN) &&
		(gamesList.firstShown + NUMBER_OF_GAMES_TO_SHOW < gamesList.count))
	{
			return 1;
	}
	else if ((direction == DIRECTION_UP) &&
	         (gamesList.firstShown))
	{
		return 1;
	}
	return 0;
}



void update_game_params()
{
}


// Print a given metastring. E.g. print_meta_string(3) to print MS3.
//
void print_meta_string(u16 msNum)
{
	int i;
	u8 *pStr = (u8*)metaStrings[msNum];
	u8 attribs;
	u16 vramOffs;

	for (;;)
	{
		if (*pStr == 0)
		{
			// We've reached the null-terminator
			break;
		}
		else if (*pStr == 0xff)
		{
			// The next three bytes contain row, column and palette number
			vramOffs = *(++pStr) << 5;
			vramOffs |= *(++pStr);
			vramOffs <<= 1;
			attribs = *(++pStr) << 2;
		}
		else
		{
			// This byte was a character
			bg0Buffer[vramOffs++] = *pStr;
			bg0Buffer[vramOffs++] = attribs;
		}
		pStr++;
	}
	bg0BufferDirty = 1;
}


void puts_game_title(u16 gameNum, u16 vramOffs, u8 attributes)
{
	u8 *pGame;
	int i;

	set_full_pointer((void**)&pGame, GAME_LIST_BANK, 0xc800 + (gameNum << 6));

	// Make sure that the chunk begins with 0xff
	//
	if (*pGame == 0xff)
	{
		pGame += 0x0c;		// Add 0x0c to the address, since that's where the actual string is located
		for (i = 0; i < 28; i++)
		{
			if (pGame[i])
			{
				bg0Buffer[vramOffs++] = pGame[i];
				bg0Buffer[vramOffs++] = attributes;
			}
			else
			{
				break;
			}
		}
	}
	bg0BufferDirty = 1;
}



// Prints info about the highlighted game (size, type, dsp etc.)
//
void print_highlighted_game_info()
{
	u8 *pGame;

	set_full_pointer((void**)&pGame, GAME_LIST_BANK, 0xc800 + (gamesList.highlighted << 6));

	// Print "GAME (nnn)"
	print_meta_string(4);

	// Print game size
	print_meta_string((pGame[2] >> 4) + 33);

	// Print save size
	print_meta_string((pGame[4] & 0xf) + 10);

	// Print ROM type
	print_meta_string(((pGame[6] >> 4) & 1) + 48);

	// Print DSP type
	print_meta_string((pGame[5] >> 4) + 50);
}


void show_scroll_indicators()
{
	// Metastring 33 and 34 are the up/down arrows. 35 and 36 are spaces with the same positions
	print_meta_string(35 - (can_games_list_scroll(DIRECTION_UP) << 1));
	print_meta_string(36 - (can_games_list_scroll(DIRECTION_DOWN) << 1));
}


void print_games_list()
{
	int i;
	u16 vramOffs = 0x0244;

	print_highlighted_game_info();
	show_scroll_indicators();

	for (i = 0; i < NUMBER_OF_GAMES_TO_SHOW; i++)
	{
		puts_game_title(gamesList.firstShown + i,
		                vramOffs,
		                (gamesList.highlighted == gamesList.firstShown + i)?8:24);
		vramOffs += 0x40;
	}

}


void hide_games_list()
{
	int i;

	// Hide the top and bottom arrows
	print_meta_string(35);
	print_meta_string(36);

	for (i = 0; i < 320; i++)
	{
		bg0Buffer[0x240 + i + i] = ' ';
		bg0Buffer[0x240 + i + i + 1] = 0;
	}
	bg0BufferDirty = 1;
}


// Clears the status window (the one showing the ROM size, save size etc.)
//
void clear_status_window()
{
	int i;

	for (i = 0; i < 96; i++)
	{
		bg0Buffer[0x540 + i + i] = ' ';
		bg0Buffer[0x540 + i + i + 1] = 0;
	}
	bg0BufferDirty = 1;
}


// Show hardware and card revision
void print_hw_card_rev()
{
	if (cardModel == 1)
	{
		print_meta_string(5);
	}
	if (cardModel == 2)
	{
		print_meta_string(6);
	}
	if (cpID == 1)
	{
		print_meta_string(9);
	}
}



// Move to the next entry in the games list
//
void move_to_next_game()
{
	if (++gamesList.highlighted >= gamesList.count)
	{
		gamesList.highlighted--;
	}
	else
	{
		update_game_number_string(1);

		// Check if the games list needs to be scrolled
		if ((gamesList.firstShown + ((NUMBER_OF_GAMES_TO_SHOW / 2) - 1) < gamesList.highlighted) &&
		    (gamesList.firstShown + NUMBER_OF_GAMES_TO_SHOW < gamesList.count))
		{
			gamesList.firstShown++;
		}
	}
	update_game_params();
	print_games_list();
}


// Move to the previous entry in the games list
//
void move_to_previous_game()
{
	if (gamesList.highlighted)
	{
		gamesList.highlighted--;

		update_game_number_string(-1);

		// Check if the games list needs to be scrolled
		if ((gamesList.firstShown) &&
		    (gamesList.firstShown + ((NUMBER_OF_GAMES_TO_SHOW / 2) - 1) >= gamesList.highlighted))
		{
			gamesList.firstShown--;
		}
	}

	update_game_params();
	print_games_list();
}


// Move to the next page in the games list (NUMBER_OF_GAMES_TO_SHOW entries down)
//
void move_to_next_page()
{
	if (gamesList.highlighted + NUMBER_OF_GAMES_TO_SHOW < gamesList.count)
	{
		gamesList.highlighted += NUMBER_OF_GAMES_TO_SHOW;
		update_game_number_string(NUMBER_OF_GAMES_TO_SHOW);

		// firstShown = highlighted - (highlighted % NUMBER_OF_GAMES_TO_SHOW)
		gamesList.firstShown = gamesList.highlighted - hw_div16_8_rem16(gamesList.highlighted, NUMBER_OF_GAMES_TO_SHOW);

		// Make sure firstShown is in range
		for (; gamesList.firstShown + NUMBER_OF_GAMES_TO_SHOW > gamesList.count; gamesList.firstShown--);
		for (; gamesList.firstShown > gamesList.highlighted; gamesList.firstShown--);
	}

	update_game_params();
	print_games_list();
}


// Move to the previous page in the games list (NUMBER_OF_GAMES_TO_SHOW entries up)
//
void move_to_previous_page()
{
	if (gamesList.highlighted >= NUMBER_OF_GAMES_TO_SHOW)
	{
		gamesList.highlighted -= NUMBER_OF_GAMES_TO_SHOW;
		update_game_number_string(-NUMBER_OF_GAMES_TO_SHOW);

		// firstShown = highlighted - (highlighted % NUMBER_OF_GAMES_TO_SHOW)
		gamesList.firstShown = gamesList.highlighted - hw_div16_8_rem16(gamesList.highlighted, NUMBER_OF_GAMES_TO_SHOW);

		// Make sure firstShown is in range
		for (; gamesList.firstShown + NUMBER_OF_GAMES_TO_SHOW > gamesList.count; gamesList.firstShown--);
		for (; gamesList.firstShown > gamesList.highlighted; gamesList.firstShown--);
	}

	update_game_params();
	print_games_list();

}


// TODO. This code needs to run from RAM
void run_game_from_gba_card_helper()
{
	u8 b;

	MYTH_GBAC_LIO = 0xe0;
	b = *(u8*)0xffa400;

	MYTH_GBAC_LIO = 0;
	b = *(u8*)0xc02a00;
	b = *(u8*)0xc3a400;
	b = *(u8*)0xc42a00;

	MYTH_GBAC_LIO = 0xe0;
	b = *(u8*)0xfc2a00;
}


// TODO. This code needs to run from RAM
void run_game_from_gba_card()
{
	u8 b, countLo, countHi, j;
	int psramParamsId, i;
	void (*copy_game_1mbit)(void);

	clear_status_window();

	MYTH_GBAC_ZIO = 0x20;					// A21 off

	run_game_from_gba_card_helper();

	MYTH_GBAC_LIO = 0x20;
	b = *(u8*)0xee2406;

	run_game_from_gba_card_helper();
	if (cardModel == 0)
	{
		MYTH_GBAC_LIO = 0xc0;
		b = *(u8*)0xf55c88;
	}
	else if (cardModel == 1)
	{
		MYTH_GBAC_LIO = 0xc0;
		b = *(u8*)0xf51c88;
	}
	else
	{
		MYTH_GBAC_LIO = 0xc0;
		b = *(u8*)0xf41c88;
	}

	run_game_from_gba_card_helper();
	MYTH_GBAC_LIO = 0xe0;
	b = *(u8*)0xdc0c60;

	MYTH_GBAC_HIO = romAddressPins >> 8;	// Set A24 & A25
	MYTH_EXTM_ON = 1;						// A24 & A25 on

	// Select the right parameters for the current ROM size
	for (psramParamsId = 0; psramParamsId < 8; psramParamsId++)
	{
		if (romSize == psramParameters[psramParamsId].romSize)
		{
			break;
		}
	}

	// The copy loop has been moved to RAM at 0x7e6000, so point our function pointer to
	// that address.
	set_full_pointer((void**)&copy_game_1mbit, 0x7e, 0x6000);

	MYTH_OPTION_IO = 0x04;					// Copy mode
	MYTH_WE_IO = 0x01;						// PSRAM WE on

	MYTH_GBAC_ZIO = 0xf8;					// GBA card 8M size
	MYTH_PRAM_ZIO = 0xf8;					// PSRAM 8M size

	MYTH_GBAC_LIO = (u8)romAddressPins;

	MYTH_PRAM_BIO = 0x00;

	// Copy the game from the GBA card to PSRAM
	countLo = psramParameters[psramParamsId].countLo;
	countHi = psramParameters[psramParamsId].countHi;
	j = 0;
	do
	{
		i = 0;
		for (; countLo; countLo--)
		{
			// Modify some of the addresses in the copy loop, which resides in RAM starting
			// at 0x7e6000
			*(u8*)0x7e6007 = 0x40 + i;	// First bank to read from
			*(u8*)0x7e6054 = 0x51 + i;	// Second bank to write to
			*(u8*)0x7e6018 = 0x50 + i;	// First bank to write to
			*(u8*)0x7e601f = 0x50 + i;	// ...
			*(u8*)0x7e6026 = 0x50 + i;
			*(u8*)0x7e602d = 0x50 + i;
			*(u8*)0x7e6034 = 0x50 + i;
			*(u8*)0x7e603b = 0x50 + i;
			*(u8*)0x7e6042 = 0x50 + i;
			*(u8*)0x7e6049 = 0x50 + i;

			copy_game_1mbit();

			i += 2;
		}

		MYTH_PRAM_BIO = j + 1;
		b = (u8)romAddressPins;
		romAddressPins &= 0xff00;
		romAddressPins |= b + 8;
		countLo = 8;
	} while (++j != countHi);

	MYTH_WE_IO = 0;						// PSRAM WE off
	MYTH_OPTION_IO = MAP_MENU_FLASH_TO_ROM;
	MYTH_GBAC_ZIO = 0x20;

	run_game_from_gba_card_helper();
	MYTH_GBAC_LIO = 0x20;
	b = *(u8*)0xee2006;

	run_game_from_gba_card_helper();
	MYTH_GBAC_LIO = 0xc0;
	b = *(u8*)0xf50088;

	MYTH_GBAC_LIO = 0;
	MYTH_GBAC_HIO = 0;
	MYTH_GBAC_ZIO = 0;

	// SET  SNES PSRAM RUN
	MYTH_OPTION_IO = 1;
	MYTH_PRAM_BIO = 0;
	MYTH_PRAM_ZIO = psramParameters[psramParamsId].romZ;
	MYTH_GBAS_BIO = 0;

	// TODO: finish the rest of the function
}


// TODO. This code needs to run from RAM
void run_secondary_cart()
{
	int i;
	u8 *pCartName;
	u16 keys;

	clear_screen();

	*(u8*)0x00c017 = 1;
	MYTH_OPTION_IO = 0x05;

	set_full_pointer((void**)&pCartName, 0x00, 0xffc0);

	// Fill in the cartridge name
	for (i = 0; i < 20; i++)
	{
		MS3[i + 4] = pCartName[i];
	}

	print_meta_string(3);				// Print cart name
	print_meta_string(58);				// Print instructions
	print_hw_card_rev();

	// Print region and CPU/PPU1/PPU2 versions
	print_meta_string(((REG_STAT78 >> 4) & 1) + 59);
	print_meta_string((REG_RDNMI & 3) + 61);
	print_meta_string((REG_STAT77 & 3) + 65);
	print_meta_string((REG_STAT78 & 3) + 69);

	dma_bg0_buffer();

	// Wait until B or Y is pressed
	while (!((keys = read_joypad()) & 0xc000));

	if (keys & 0x8000)
	{
		// B
		*(u8*)0x00c017 = 0;
		MYTH_OPTION_IO = 0;

		// Reset the "GAME (nnn)" string
		//MS4[0xb] = MS4[0xc] = '0';
		//MS4[0xd] = '1';

		clear_screen();
		print_meta_string(0);
		print_meta_string(4);
		print_hw_card_rev();
		print_games_list();
	}
	else if (keys & 0x4000)
	{
		// Y
		*(u8*)0x00c017 = 0;
		MYTH_OPTION_IO = 0x05;
		run_3800();
	}
}



int main()
{
	u8 *bp;
	int i;
	u16 keys;

	*(u8*)0x00c040 = 0;				// CPLD ROM

	if ((*(u8*)0x00ffe1 == 0x63) &&
	    (*(u8*)0x00ffe2 == 0x57))
	{
		cpID = *(u8*)0x00ffe0;
	}

	*(u8*)0x00c040 = 1;				// GBA card menu

	cardModel = *(u8*)0x00fff0;

	gamesList.highlighted = gamesList.firstShown = 0;
	gamesList.count = count_games_on_gba_card();
	create_alphabetical_index();

	copy_ram_code();

	REG_DISPCNT = 0x80;				// Turn screen off

	expand_font_data();
	load_font_colors();

	// Load the background graphics data
	lzss_decode_vram(&bg_patterns, 0x4000, (&bg_patterns_end - &bg_patterns));
  	load_cgram(&bg_palette, 0, 32);
   	load_vram(&bg_map, 0x3000,(&bg_map_end - &bg_map));

	// Setup various PPU registers
	bp = &(REG_OBSEL);
	for (i = 0; i < 12; i++)
	{
		*bp++ = ppuRegData1[i];
	}
	bp = &(REG_BGCNT);
	for (i = 0; i < 9; i++)
	{
		*bp++ = ppuRegData2[i];
	}

	// Setup color addition parameters
	REG_COLDATA = 0x22;
	REG_COLDATA = 0x41;
	REG_COLDATA = 0x82;

	REG_DISPCNT = 0x0f;			// Screen on, full brightness

	// Reset scrolling for BG0 and BG1
	REG_BG0VOFS = REG_BG0HOFS = 0;
	REG_BG0VOFS = REG_BG0HOFS = 0;
	REG_BG1VOFS = REG_BG1HOFS = 0;
	REG_BG1VOFS = REG_BG1HOFS = 0;

	REG_NMI_TIMEN = 1;

	// Clear the BG0 RAM buffer
	clear_screen();

	print_meta_string(0);
	print_meta_string(4);
	print_hw_card_rev();
	print_games_list();

	while (1)
	{
		if (bg0BufferDirty)
		{
			dma_bg0_buffer();
		}

		keys = read_joypad();

		if (keys & 0x8000)
		{
			// B
			run_game_from_gba_card();
		}
		else if (keys & 0x4000)
		{
			// Y
			run_secondary_cart();
		}
		else if (keys & 0x0800)
		{
			// Up
			move_to_previous_game();
		}
		else if (keys & 0x0400)
		{
			// Down
			move_to_next_game();
		}
		else if (keys & 0x0010)
		{
			// R
			move_to_next_page();
		}
		else if (keys & 0x0020)
		{
			// L
			move_to_previous_page();
		}
	}


	return 0;
}

