; Various support code in assembly for the SNES Myth Menu
; Neo TeAm, 2010
; /Mic, 2010

.include "hdr.asm"
.include "snes_io.inc"
.include "myth_io.inc"


.section ".text_neo2" superfree

; Divide a 16-bit number by an 8-bit number and return the 16-bit quotient
hw_div16_8_quot16:
	sep	#$20
	lda	4,s
	sta.l 	$4204   	; Dividend low
	lda 	5,s
	sta.l 	$4205 		; Dividend high
	lda 	6,s
	sta.l 	$4206
	nop			; Wait for the division to finish
	nop			; ...
	nop	
	nop
	nop
	nop
	nop
	nop
	rep	#$20
	lda.l	$4214		; Quotient
	sta.b 	tcc__r0		; Store return value
	rtl
	

; Divide a 16-bit number by an 8-bit number and return the 16-bit remainder	
hw_div16_8_rem16:
	sep	#$20
	lda	4,s
	sta.l 	$4204   	; Dividend low
	lda 	5,s
	sta.l 	$4205 		; Dividend high
	lda 	6,s
	sta.l 	$4206
	nop			; Wait for the division to finish
	nop			; ...
	nop	
	nop
	nop
	nop
	nop
	nop
	rep	#$20
	lda.l	$4216		; Remainder
	sta.b 	tcc__r0		; Store return value
	rtl

; Divide a 16-bit number by an 8-bit number and return the 8-bit quotient and 8-bit remainder
hw_div16_8_quot8_rem8:
	sep	#$20
	lda	4,s
	sta.l 	$4204   	; Dividend low
	lda 	5,s
	sta.l 	$4205 		; Dividend high
	lda 	6,s
	sta.l 	$4206
	nop			; Wait for the division to finish
	nop			; ...
	nop	
	nop
	nop
	nop
	nop
	nop
	lda.l	$4216		; Remainder
	xba			; put remainder in high byte
	lda.l	$4214		; Quotient in low byte
	rep	#$20
	sta.b 	tcc__r0		; Store return value
	rtl


; Used for running the secondary cart (plugged in at the back of the Myth)
run_3800:
        LDA     #$01                 ;
        STA.L   $C00E                ; HARD REST IO

        LDA     #$01                 ;
        STA.L   $C012                ; CPLD RAM ON
        STA.L   $C015                ; CPLD RAM ON
              
        LDA     #$FF                 ; RUN
        STA.L   MYTH_RST_IO          ; RUN
        
        LDA     #$0   ;#$00 0000     ;MYTH
        PHA                          ;
        PLB
         
        LDX     #$00                 ;0  MOVE BOOT CODE TO SRAM
-:      LDA.L   CPLD_RAM+$7E0000,X  ;0
        STA.L   $3800,X             ;0  CPLD SRAM
        INX
        BNE     -                   ;0

        LDX     #$00                 ;0  MOVE CHEAT CODE TO SRAM
-:      LDA.L   CHEAT+$7E0000,X     ;0
        STA.L   $3E00,X             ;0  CPLD SRAM
        INX
        BNE     -                   ;0
        
        JMP.L   $003800
;===============================================================================
; EXIT RAM  $003800~003FFF
;===============================================================================
CPLD_RAM:
        REP	#$30        ; A,8 & X,Y 16 BIT
        LDX     #$0000
-:
        LDA     #$0000      ; READ GBA CARD
        STA.L   $7E0000,X  ; CLERA SRAM
        INX
        INX
        BNE     -
        SEP	#$30

        LDA     #$00                 ;
        STA.L   REG_DISPCNT
        STA.L   REG_MDMAEN
        STA.L   REG_HDMAEN
        STA.L   REG_NMI_TIMEN
        SEI
        SEC
        XCE              ;SET 65C02 MODE
        JMP     ($FFFC)

CHEAT:
        LDA   #$63
        STA.L $7EFA91
        
        LDA   #$01
-:
        BIT   $4012
        BNE   -
        RTL
        
        
ram_code_begin:

;===============================================================================
; Copy 1 Mbit of data from the GBA Card to PSRAM
;
; DBR is set to the source bank ($4x) so that word addressing can be used
; when reading data from the GBA Card, thus saving 1 cycle per read.
;
; Then 512 kb (64 kB) are copied at 16 bytes per iteration. The destination bank
; is then incremented by overwriting parts of the code, and then another 512 kb
; are copied.
copy_1mbit_from_gbac_to_psram:
	PHP
    	SEP	#$20
    	REP	#$10
    	PHB        	   ; Save DBR
    	LDA   	#$40	   ; +$07	
	LDY	#2	   ; Copy 2 * 512 kbit	
MOV_512K:	
    	PHA
    	PLB        	   ; Set DBR

    	REP	#$30       ; A,16 & X,Y 16 BIT
    	LDX     #$1FFE
-:
    	LDA.W   $0000,X    ; READ GBA CARD
    	STA.L   $500000,X  ; WRITE PSRAM ;+$18
    	LDA.W   $2000,X    ; READ GBA CARD
    	STA.L   $502000,X  ; WRITE PSRAM 
    	LDA.W   $4000,X    ; ...
    	STA.L   $504000,X 
    	LDA.W   $6000,X   
    	STA.L   $506000,X 
    	LDA.W   $8000,X   
    	STA.L   $508000,X 
    	LDA.W   $A000,X    
    	STA.L   $50A000,X 
    	LDA.W   $C000,X   
    	STA.L   $50C000,X  
    	LDA.W   $E000,X   
    	STA.L   $50E000,X  
    	DEX
    	DEX
    	BPL     -

	DEY
	BEQ	+
;+$51	
	SEP	#$20
	LDA	#$51		;+$54
	STA.L	$7e6000+$18	; Increase the MSB of all the addresses above on the form $5xxxxx
	STA.L	$7e6000+$1F	; ...
	STA.L	$7e6000+$26
	STA.L	$7e6000+$2D
	STA.L	$7e6000+$34
	STA.L	$7e6000+$3B
	STA.L	$7e6000+$42
	STA.L	$7e6000+$49
	
	; Increase DBR
	PHB
	PLA
	INA
	BRA	MOV_512K
+:	
    	PLB        	   ; Restore DBR
    	PLP
    	RTS
    	

ram_code_end:


copy_ram_code:
	php
	rep	#$10
	sep	#$20
	ldy	#(ram_code_end - ram_code_begin)
	ldx	#0
-:	
	lda.l	ram_code_begin,x
	sta.l	$7e6000,x
	inx
	dey
	bne	-
	plp
	rtl
	
	
.ends


